rem java org.antlr.v4.Tool -Dlanguage=Python2 -long-messages aml.g4 -o ./py2/
java org.antlr.v4.Tool -Dlanguage=Python3 -long-messages aml.g4 -o .
rem java org.antlr.v4.Tool -Dlanguage=Python2 -long-messages a2l.g4 -o ./py2/
java org.antlr.v4.Tool -Dlanguage=Python3 -long-messages a2l.g4 -o .
