/*
**
**  !!! AUTOMATICALLY GENERATED FILE -- DO NOT EDIT !!!
**
*/
#if !defined(__A2LTOKEN_HPP)
    #define __A2LTOKEN_HPP

    #include <cstdint>
    #include <map>
    #include <string>

enum class A2LTokenType : uint16_t {
    INVALID                      = 0,
    ALIGNMENT_BYTE               = 1,
    ALIGNMENT_FLOAT16_IEEE       = 2,
    ALIGNMENT_FLOAT32_IEEE       = 3,
    ALIGNMENT_FLOAT64_IEEE       = 4,
    ALIGNMENT_INT64              = 5,
    ALIGNMENT_LONG               = 6,
    ALIGNMENT_WORD               = 7,
    ANNOTATION                   = 8,
    ANNOTATION_LABEL             = 9,
    ANNOTATION_ORIGIN            = 10,
    ANNOTATION_TEXT              = 11,
    BIT_MASK                     = 12,
    _BYTE_ORDER                  = 13,
    CALIBRATION_ACCESS           = 14,
    CALIBRATION                  = 15,
    NO_CALIBRATION               = 16,
    NOT_IN_MCD_SYSTEM            = 17,
    OFFLINE_CALIBRATION          = 18,
    DEFAULT_VALUE                = 19,
    DEPOSIT                      = 20,
    ABSOLUTE                     = 21,
    DIFFERENCE                   = 22,
    DISCRETE                     = 23,
    DISPLAY_IDENTIFIER           = 24,
    ECU_ADDRESS_EXTENSION        = 25,
    EXTENDED_LIMITS              = 26,
    FORMAT                       = 27,
    FUNCTION_LIST                = 28,
    GUARD_RAILS                  = 29,
    IF_DATA                      = 30,
    MATRIX_DIM                   = 31,
    MAX_REFRESH                  = 32,
    MONOTONY                     = 33,
    MON_DECREASE                 = 34,
    MON_INCREASE                 = 35,
    STRICT_DECREASE              = 36,
    STRICT_INCREASE              = 37,
    MONOTONOUS                   = 38,
    STRICT_MON                   = 39,
    NOT_MON                      = 40,
    PHYS_UNIT                    = 41,
    READ_ONLY                    = 42,
    REF_CHARACTERISTIC           = 43,
    REF_MEMORY_SEGMENT           = 44,
    REF_UNIT                     = 45,
    STEP_SIZE                    = 46,
    SYMBOL_LINK                  = 47,
    VERSION                      = 48,
    ASAP2_VERSION                = 49,
    A2ML_VERSION                 = 50,
    PROJECT                      = 51,
    HEADER                       = 52,
    PROJECT_NO                   = 53,
    MODULE                       = 54,
    A2ML                         = 55,
    AXIS_PTS                     = 56,
    CHARACTERISTIC               = 57,
    ASCII                        = 58,
    CURVE                        = 59,
    MAP                          = 60,
    CUBOID                       = 61,
    CUBE_4                       = 62,
    CUBE_5                       = 63,
    VAL_BLK                      = 64,
    VALUE                        = 65,
    AXIS_DESCR                   = 66,
    CURVE_AXIS                   = 67,
    COM_AXIS                     = 68,
    FIX_AXIS                     = 69,
    RES_AXIS                     = 70,
    STD_AXIS                     = 71,
    AXIS_PTS_REF                 = 72,
    CURVE_AXIS_REF               = 73,
    FIX_AXIS_PAR                 = 74,
    FIX_AXIS_PAR_DIST            = 75,
    FIX_AXIS_PAR_LIST            = 76,
    MAX_GRAD                     = 77,
    COMPARISON_QUANTITY          = 78,
    DEPENDENT_CHARACTERISTIC     = 79,
    MAP_LIST                     = 80,
    NUMBER                       = 81,
    VIRTUAL_CHARACTERISTIC       = 82,
    COMPU_METHOD                 = 83,
    IDENTICAL                    = 84,
    FORM                         = 85,
    LINEAR                       = 86,
    RAT_FUNC                     = 87,
    TAB_INTP                     = 88,
    TAB_NOINTP                   = 89,
    TAB_VERB                     = 90,
    COEFFS                       = 91,
    COEFFS_LINEAR                = 92,
    COMPU_TAB_REF                = 93,
    FORMULA                      = 94,
    FORMULA_INV                  = 95,
    STATUS_STRING_REF            = 96,
    COMPU_TAB                    = 97,
    DEFAULT_VALUE_NUMERIC        = 98,
    COMPU_VTAB                   = 99,
    COMPU_VTAB_RANGE             = 100,
    FRAME                        = 101,
    FRAME_MEASUREMENT            = 102,
    FUNCTION                     = 103,
    DEF_CHARACTERISTIC           = 104,
    FUNCTION_VERSION             = 105,
    IN_MEASUREMENT               = 106,
    LOC_MEASUREMENT              = 107,
    OUT_MEASUREMENT              = 108,
    SUB_FUNCTION                 = 109,
    GROUP                        = 110,
    REF_MEASUREMENT              = 111,
    ROOT                         = 112,
    SUB_GROUP                    = 113,
    INSTANCE                     = 114,
    MEASUREMENT                  = 115,
    ARRAY_SIZE                   = 116,
    BIT_OPERATION                = 117,
    LEFT_SHIFT                   = 118,
    RIGHT_SHIFT                  = 119,
    SIGN_EXTEND                  = 120,
    ECU_ADDRESS                  = 121,
    ERROR_MASK                   = 122,
    LAYOUT                       = 123,
    ROW_DIR                      = 124,
    COLUMN_DIR                   = 125,
    READ_WRITE                   = 126,
    VIRTUAL                      = 127,
    MOD_COMMON                   = 128,
    DATA_SIZE                    = 129,
    S_REC_LAYOUT                 = 130,
    MOD_PAR                      = 131,
    ADDR_EPK                     = 132,
    CALIBRATION_METHOD           = 133,
    CALIBRATION_HANDLE           = 134,
    CALIBRATION_HANDLE_TEXT      = 135,
    CPU_TYPE                     = 136,
    CUSTOMER                     = 137,
    CUSTOMER_NO                  = 138,
    ECU                          = 139,
    ECU_CALIBRATION_OFFSET       = 140,
    EPK                          = 141,
    MEMORY_LAYOUT                = 142,
    PRG_CODE                     = 143,
    PRG_DATA                     = 144,
    PRG_RESERVED                 = 145,
    MEMORY_SEGMENT               = 146,
    CALIBRATION_VARIABLES        = 147,
    CODE                         = 148,
    DATA                         = 149,
    EXCLUDE_FROM_FLASH           = 150,
    OFFLINE_DATA                 = 151,
    RESERVED                     = 152,
    SERAM                        = 153,
    VARIABLES                    = 154,
    EEPROM                       = 155,
    EPROM                        = 156,
    FLASH                        = 157,
    RAM                          = 158,
    ROM                          = 159,
    REGISTER                     = 160,
    INTERN                       = 161,
    EXTERN                       = 162,
    NO_OF_INTERFACES             = 163,
    PHONE_NO                     = 164,
    SUPPLIER                     = 165,
    SYSTEM_CONSTANT              = 166,
    USER                         = 167,
    RECORD_LAYOUT                = 168,
    AXIS_PTS_X                   = 169,
    AXIS_PTS_Y                   = 170,
    AXIS_PTS_Z                   = 171,
    AXIS_PTS_4                   = 172,
    AXIS_PTS_5                   = 173,
    AXIS_RESCALE_X               = 174,
    AXIS_RESCALE_Y               = 175,
    AXIS_RESCALE_Z               = 176,
    AXIS_RESCALE_4               = 177,
    AXIS_RESCALE_5               = 178,
    DIST_OP_X                    = 179,
    DIST_OP_Y                    = 180,
    DIST_OP_Z                    = 181,
    DIST_OP_4                    = 182,
    DIST_OP_5                    = 183,
    FIX_NO_AXIS_PTS_X            = 184,
    FIX_NO_AXIS_PTS_Y            = 185,
    FIX_NO_AXIS_PTS_Z            = 186,
    FIX_NO_AXIS_PTS_4            = 187,
    FIX_NO_AXIS_PTS_5            = 188,
    FNC_VALUES                   = 189,
    ALTERNATE_CURVES             = 190,
    ALTERNATE_WITH_X             = 191,
    ALTERNATE_WITH_Y             = 192,
    IDENTIFICATION               = 193,
    NO_AXIS_PTS_X                = 194,
    NO_AXIS_PTS_Y                = 195,
    NO_AXIS_PTS_Z                = 196,
    NO_AXIS_PTS_4                = 197,
    NO_AXIS_PTS_5                = 198,
    STATIC_RECORD_LAYOUT         = 199,
    NO_RESCALE_X                 = 200,
    NO_RESCALE_Y                 = 201,
    NO_RESCALE_Z                 = 202,
    NO_RESCALE_4                 = 203,
    NO_RESCALE_5                 = 204,
    OFFSET_X                     = 205,
    OFFSET_Y                     = 206,
    OFFSET_Z                     = 207,
    OFFSET_4                     = 208,
    OFFSET_5                     = 209,
    RIP_ADDR_W                   = 210,
    RIP_ADDR_X                   = 211,
    RIP_ADDR_Y                   = 212,
    RIP_ADDR_Z                   = 213,
    RIP_ADDR_4                   = 214,
    RIP_ADDR_5                   = 215,
    SHIFT_OP_X                   = 216,
    SHIFT_OP_Y                   = 217,
    SHIFT_OP_Z                   = 218,
    SHIFT_OP_4                   = 219,
    SHIFT_OP_5                   = 220,
    SRC_ADDR_X                   = 221,
    SRC_ADDR_Y                   = 222,
    SRC_ADDR_Z                   = 223,
    SRC_ADDR_4                   = 224,
    SRC_ADDR_5                   = 225,
    TYPEDEF_CHARACTERISTIC       = 226,
    TYPEDEF_MEASUREMENT          = 227,
    TYPEDEF_STRUCTURE            = 228,
    STRUCTURE_COMPONENT          = 229,
    UNIT                         = 230,
    DERIVED                      = 231,
    EXTENDED_SI                  = 232,
    SI_EXPONENTS                 = 233,
    UNIT_CONVERSION              = 234,
    USER_RIGHTS                  = 235,
    REF_GROUP                    = 236,
    VARIANT_CODING               = 237,
    VAR_CHARACTERISTIC           = 238,
    VAR_ADDRESS                  = 239,
    VAR_CRITERION                = 240,
    VAR_MEASUREMENT              = 241,
    VAR_SELECTION_CHARACTERISTIC = 242,
    VAR_FORBIDDEN_COMB           = 243,
    VAR_NAMING                   = 244,
    NUMERIC                      = 245,
    APLHA                        = 246,
    VAR_SEPARATOR                = 247,
    DOT                          = 248,
    BRO                          = 249,
    BRC                          = 250,
    UBYTE                        = 251,
    SBYTE                        = 252,
    UWORD                        = 253,
    SWORD                        = 254,
    ULONG                        = 255,
    SLONG                        = 256,
    A_UINT64                     = 257,
    A_INT64                      = 258,
    FLOAT16_IEEE                 = 259,
    FLOAT32_IEEE                 = 260,
    FLOAT64_IEEE                 = 261,
    BYTE                         = 262,
    WORD                         = 263,
    LONG                         = 264,
    PBYTE                        = 265,
    PWORD                        = 266,
    PLONG                        = 267,
    DIRECT                       = 268,
    _LITTLE_ENDIAN               = 269,
    _BIG_ENDIAN                  = 270,
    MSB_LAST                     = 271,
    MSB_FIRST                    = 272,
    INDEX_INCR                   = 273,
    INDEX_DECR                   = 274,
    SYMBOL_TYPE_LINK             = 275,
    BEGIN                        = 276,
    END                          = 277,
    IDENT                        = 278,
    FLOAT                        = 279,
    INT                          = 280,
    HEX                          = 281,
    COMMENT                      = 282,
    WS                           = 283,
    STRING                       = 284,
    // Ext.
    NOT_IN_ECU			 = 290,
    TYPEDEF_AXIS                 = 291,
    MODEL_LINK                   = 292,
    TRANSFORMER			 = 293,
    TRANSFORMER_IN_OBJECTS	 = 294,
    TRANSFORMER_OUT_OBJECTS	 = 295,
    BLOB			 = 296,
};

#endif  // __A2LTOKEN_HPP
